import logging
import greeclimate.network_helper as nethelper

from enum import IntEnum, unique
from greeclimate.exceptions import DeviceNotBoundError
from greeclimate.network_helper import Props


@unique
class TemperatureUnits(IntEnum):
    C = 0
    F = 1


@unique
class Mode(IntEnum):
    Auto = 0
    Cool = 1
    Dry = 2
    Fan = 3
    Heat = 4


@unique
class FanSpeed(IntEnum):
    Auto = 0
    Low = 1
    MediumLow = 2
    Medium = 3
    MediumHigh = 4
    High = 5


@unique
class HorizontalSwing(IntEnum):
    Default = 0
    FullSwing = 1
    Left = 2
    LeftCenter = 3
    Center = 4
    RightCenter = 5
    Right = 6


@unique
class VerticalSwing(IntEnum):
    Default = 0
    FullSwing = 1
    FixedUpper = 2
    FixedUpperMiddle = 3
    FixedMiddle = 4
    FixedLowerMiddle = 5
    FixedLower = 6
    SwingUpper = 7
    SwingUpperMiddle = 8
    SwingMiddle = 9
    SwingLowerMiddle = 10
    SwingLower = 11


class Device:
    """Class representing a physical device, it's state and properties.

    Devices must be bound, either by discovering their presence, or supplying a persistent
    device key which is then used for communication (and encryption) with the unit. See the
    `bind` function for more details on how to do this.

    Once a device is bound occasionally call `update_state` to request and update state from
    the HVAC, as it is possible that it changes state from other sources.

    Attributes:
        power: A boolean indicating if the unit is on or off
        mode: An int indicating operating mode, see `Mode` enum for possible values
        target_temperature: The target temperature, ignore if in Auto, Fan or Steady Heat mode
        temperature_units: An int indicating unit of measurement, see `TemperatureUnits` enum for possible values
        fan_speed: An int indicating fan speed, see `FanSpeed` enum for possible values
        fresh_air: A boolean indicating if fresh air valve is open, if present
        xfan: A boolean to enable the fan to dry the coil, only used for cool and dry modes
        anion: A boolean to enable the ozone generator, if present
        sleep: A boolean to enable sleep mode, which adjusts temperature over time
        light: A boolean to enable the light on the unit, if present
        horizontal_swing: An int to control the horizontal blade position, see `HorizontalSwing` enum for possible values
        vertical_swing: An int to control the vertical blade position, see `VerticalSwing` enum for possible values
        quiet: A boolean to enable quiet operation
        turbo: A boolean to enable turbo operation (heat or cool faster initially)
        steady_heat: When enabled unit will maintain a target temperature of 8 degrees C
        power_save: A boolen to enable power save operation
    """

    def __init__(self, device_info):
        self._logger = logging.getLogger(__name__)

        self.device_info = device_info
        self.device_key = None

        """ Device properties """
        self._properties = None

    async def bind(self, key=None):
        """ Run the binding procedure.
        
        Binding is a finnicky procedure, and happens in 1 of 2 ways:
            1 - Without the key, binding must pass the device info structure immediately following
                the search devices procedure. There is only a small window to complete registration.
            2 - With a key, binding is implicit and no further action is required

            Both approaches result in a device_key which is used as like a persitent session id.

        Raises:
            socket.timeout: If binding was unsuccessful (the device didn't respond.)
        """

        self._logger.info("Starting device binding to %s", str(self.device_info))

        if key:
            self.device_key = key
        else:
            self.device_key = await nethelper.bind_device(self.device_info)

        if self.device_key:
            self._logger.info("Bound to device using key %s", self.device_key)

    async def update_state(self):
        """ Update the internal state of the device structure of the physical device """
        if not self.device_key:
            raise DeviceNotBoundError

        self._logger.debug("Updating device properties for (%s)", str(self.device_info))

        props = [x.value for x in Props]
        self._properties = nethelper.request_state(
            props, self.device_info, self.device_key
        )

    def get_property(self, name):
        """ Generic lookup of properties tracked from the physical device """
        if self._properties:
            return self._properties.get(name.value)
        return None

    def set_property(self, name, value):
        """ Generic setting of properties for the physical device """
        if not self._properties:
            self._properties = {}

        if self._properties.get(name.value) == value:
            return
        else:
            self._properties[name.value] = value

        self._logger.debug("Sending remote state update %s -> %s", name, value)
        nethelper.send_state({name.value: value}, self.device_info, key=self.device_key)

    @property
    def power(self) -> bool:
        return bool(self.get_property(Props.POWER))

    @power.setter
    def power(self, value):
        self.set_property(Props.POWER, int(value))

    @property
    def mode(self) -> int:
        return int(self.get_property(Props.MODE))

    @mode.setter
    def mode(self, value):
        self.set_property(Props.MODE, int(value))

    @property
    def target_temperature(self) -> int:
        return int(self.get_property(Props.TEMP_SET))

    @target_temperature.setter
    def target_temperature(self, value):
        self.set_property(Props.TEMP_SET, int(value))

    @property
    def temperature_units(self) -> int:
        return int(self.get_property(Props.TEMP_UNIT))

    @temperature_units.setter
    def temperature_units(self, value):
        self.set_property(Props.TEMP_UNIT, int(value))

    @property
    def fan_speed(self) -> int:
        return int(self.get_property(Props.FAN_SPEED))

    @fan_speed.setter
    def fan_speed(self, value):
        self.set_property(Props.FAN_SPEED, int(value))

    @property
    def fresh_air(self) -> bool:
        return bool(self.get_property(Props.FRESH_AIR))

    @fresh_air.setter
    def fresh_air(self, value):
        self.set_property(Props.FRESH_AIR, int(value))

    @property
    def xfan(self) -> bool:
        return bool(self.get_property(Props.XFAN))

    @xfan.setter
    def xfan(self, value):
        self.set_property(Props.XFAN, int(value))

    @property
    def anion(self) -> bool:
        return bool(self.get_property(Props.ANION))

    @anion.setter
    def anion(self, value):
        self.set_property(Props.ANION, int(value))

    @property
    def sleep(self) -> bool:
        return bool(self.get_property(Props.SLEEP))

    @sleep.setter
    def sleep(self, value):
        self.set_property(Props.SLEEP, int(value))

    @property
    def light(self) -> bool:
        return bool(self.get_property(Props.LIGHT))

    @light.setter
    def light(self, value):
        self.set_property(Props.LIGHT, int(value))

    @property
    def horizontal_swing(self) -> int:
        return int(self.get_property(Props.SWING_HORIZ))

    @horizontal_swing.setter
    def horizontal_swing(self, value):
        self.set_property(Props.SWING_HORIZ, int(value))

    @property
    def vertical_swing(self) -> int:
        return int(self.get_property(Props.SWING_VERT))

    @vertical_swing.setter
    def vertical_swing(self, value):
        self.set_property(Props.SWING_VERT, int(value))

    @property
    def quiet(self) -> bool:
        return bool(self.get_property(Props.QUIET))

    @quiet.setter
    def quiet(self, value):
        self.set_property(Props.QUIET, int(value))

    @property
    def turbo(self) -> bool:
        return bool(self.get_property(Props.TURBO))

    @turbo.setter
    def turbo(self, value):
        self.set_property(Props.TURBO, int(value))

    @property
    def steady_heat(self) -> bool:
        return bool(self.get_property(Props.STEADY_HEAT))

    @steady_heat.setter
    def steady_heat(self, value):
        self.set_property(Props.STEADY_HEAT, int(value))

    @property
    def power_save(self) -> bool:
        return bool(self.get_property(Props.POWER_SAVE))

    @power_save.setter
    def power_save(self, value):
        self.set_property(Props.POWER_SAVE, int(value))




import asyncio
import base64
import json
import logging
import socket
import enum

from Crypto.Cipher import AES
from ipaddress import IPv4Network

from greeclimate.device_info import DeviceInfo

GENERIC_KEY = "a3K8Bx%2r8Y7#xDh"


class Props(enum.Enum):
    POWER = "Pow"
    MODE = "Mod"
    TEMP_SET = "SetTem"
    TEMP_UNIT = "TemUn"
    TEMP_BIT = "TemRec"
    FAN_SPEED = "WdSpd"
    FRESH_AIR = "Air"
    XFAN = "Blo"
    ANION = "Health"
    SLEEP = "SwhSlp"
    LIGHT = "Lig"
    SWING_HORIZ = "SwingLfRig"
    SWING_VERT = "SwUpDn"
    QUIET = "Quiet"
    TURBO = "Tur"
    STEADY_HEAT = "StHt"
    POWER_SAVE = "SvSt"
    UNKNOWN_HEATCOOLTYPE = "HeatCoolType"


def _get_broadcast_addresses():
    import netifaces

    broadcastAddrs = []

    interfaces = netifaces.interfaces()
    for iface in interfaces:
        addr = netifaces.ifaddresses(iface)
        if netifaces.AF_INET in addr:
            netmask = addr[netifaces.AF_INET][0]["netmask"]
            ipaddr = addr[netifaces.AF_INET][0]["addr"]
            net = IPv4Network(f"{ipaddr}/{netmask}", strict=False)
            if net.broadcast_address and not net.is_loopback:
                broadcastAddrs.append(str(net.broadcast_address))

    return broadcastAddrs


async def _search_on_interface(bcast, timeout):
    logger = logging.getLogger(__name__)
    logger.debug("Listening for devices on %s", bcast)

    s = create_socket(timeout)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    payload = {'t': 'scan'}
    s.sendto(json.dumps(payload).encode(), (bcast, 7000))

    devices = []
    while True:
        try:
            (d, addr) = s.recvfrom(2048)
            if (len(d)) == 0:
                continue

            response = json.loads(d)
            pack = decrypt_payload(response['pack'])
            logger.debug("Received response from device search\n%s", pack)
            devices.append(DeviceInfo(
                addr[0], addr[1], pack['cid'], pack['name']))
        except socket.timeout:
            """ Intentionally unprocessed exception """
            break
        except json.JSONDecodeError:
            logger.debug("Unable to decode device search response payload")
        except Exception as e:
            logging.error(
                "Unable to search devices due to an exception %s", str(e))

    s.close()
    return devices


async def search_devices(timeout=10, broadcastAddrs=None):
    if not broadcastAddrs:
        broadcastAddrs = _get_broadcast_addresses()

    broadcastAddrs = list(broadcastAddrs)
    done, pending = await asyncio.wait(
        [_search_on_interface(b, timeout=timeout) for b in broadcastAddrs])

    devices = []
    for task in done:
        results = task.result()
        for result in results:
            devices.append(result)

    return devices


async def bind_device(device_info):
    payload = {
        "cid": "app",
        "i": "1",
        "t": "pack",
        "uid": 0,
        "tcid": device_info.mac,
        "pack": {
            "mac": device_info.mac,
            "t": "bind",
            "uid": 0
        }
    }

    s = create_socket()
    try:
        send_data(s, device_info.ip, device_info.port, payload)
        r = receive_data(s)
    except Exception as e:
        raise e
    finally:
        s.close()

    if r["pack"]["t"] == "bindok":
        return r["pack"]["key"]


def send_state(property_values, device_info, key=GENERIC_KEY):
    payload = {
        "cid": "app",
        "i": 0,
        "t": "pack",
        "uid": 0,
        "tcid": device_info.mac,
        "pack": {
            "opt": list(property_values.keys()),
            "p": list(property_values.values()),
            "t": "cmd"
        }
    }

    s = create_socket()
    try:
        send_data(s, device_info.ip, device_info.port, payload, key=key)
        r = receive_data(s, key=key)
    except Exception as e:
        raise e
    finally:
        s.close()

    cols = r["pack"]["opt"]
    dat = r["pack"]["val"]
    return dict(zip(cols, dat))


def request_state(properties, device_info, key=GENERIC_KEY):
    payload = {
        "cid": "app",
        "i": 0,
        "t": "pack",
        "uid": 0,
        "tcid": device_info.mac,
        "pack": {
            "mac": device_info.mac,
            "t": "status",
            "cols": list(properties)
        }
    }

    s = create_socket()
    try:
        send_data(s, device_info.ip, device_info.port, payload, key=key)
        r = receive_data(s, key=key)
    except Exception as e:
        raise e
    finally:
        s.close()

    cols = r["pack"]["cols"]
    dat = r["pack"]["dat"]
    return dict(zip(cols, dat))


def decrypt_payload(payload, key=GENERIC_KEY):
    cipher = AES.new(key.encode(), AES.MODE_ECB)
    decoded = base64.b64decode(payload)
    decrypted = cipher.decrypt(decoded).decode()
    t = decrypted.replace(decrypted[decrypted.rindex('}')+1:], '')
    return json.loads(t)


def encrypt_payload(payload, key=GENERIC_KEY):
    def pad(s):
        bs = 16
        return s + (bs - len(s) % bs) * chr(bs - len(s) % bs)

    cipher = AES.new(key.encode(), AES.MODE_ECB)
    encrypted = cipher.encrypt(pad(json.dumps(payload)).encode())
    encoded = base64.b64encode(encrypted).decode()
    return encoded


def create_socket(timeout=60):
    s = socket.socket(type=socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP)
    s.settimeout(timeout)
    return s


def send_data(socket, ip, port, payload, key=GENERIC_KEY):
    payload["pack"] = encrypt_payload(payload["pack"], key)
    d = json.dumps(payload).encode()
    socket.sendto(d, (ip, int(port)))


def receive_data(socket, key=GENERIC_KEY):
    d = socket.recv(2048)
    payload = json.loads(d)
    payload["pack"] = decrypt_payload(payload["pack"], key)
    return payload



import unittest

from greeclimate.gree_climate import GreeClimate
from greeclimate.device_info import DeviceInfo
from unittest.mock import patch


class GreeClimateTestCase(unittest.IsolatedAsyncioTestCase):

    @patch("greeclimate.network_helper.search_devices")
    async def testShouldListDevicesWhenSearched(self, mock_search_devices):
        gree = GreeClimate()
        mock_search_devices.return_value = [
            DeviceInfo("1.1.1.0", "7000", "aabbcc001122", "MockDevice1")
        ]

        devices = await gree.search_devices()

        self.assertIsNotNone(devices)
        self.assertGreater(len(devices), 0, "No devices were discovered")

    @patch("greeclimate.network_helper.search_devices")
    async def testShouldReturnEmptyListDevicesWhenNoDevicesFound(self, mock_search_devices):
        gree = GreeClimate()
        mock_search_devices.return_value = []

        devices = await gree.search_devices()

        self.assertIsNotNone(devices)
        self.assertEqual(len(devices), 0)





class DeviceInfo:
    """Device information class, used to identify and connect

    Attributes
        ip: IP address (ipv4 only) of the physical device
        port: Usually this will always be 7000
        mac: mac address, in the format 'aabbcc112233'
        name: Name of unit, if available
    """
    ip = None
    port = None
    mac = None
    name = "<Unknown>"

    def __init__(self, ip, port, mac, name):
        self.ip = ip
        self.port = port
        self.mac = mac
        self.name = name if name else self.name

    def __str__(self):
        return f"Device: {self.name}, {self.ip}:{self.port} ({self.mac})"



import logging

from greeclimate.device_info import DeviceInfo
import greeclimate.network_helper as nethelper
from typing import List


class GreeClimate:
    """Interact with gree devices on the network

    The `GreeClimate` class provides basic services for discovery and interaction
    with gree device on the network.
    """

    _logger = None

    def __init__(self):
        self._logger = logging.getLogger(__name__)

    async def search_devices(self) -> List[DeviceInfo]:
        """ Sends a discovery broadcast packet on each network interface to
            locate Gree units on the network

        Returns:
            List[DeviceInfo]: List of device informations
        """
        self._logger.info("Starting Gree device discovery process")

        devices = await nethelper.search_devices()
        for d in devices:
            self._logger.info("Found %s", str(d))

        return devices



import unittest
import socket

from greeclimate.gree_climate import GreeClimate
from greeclimate.device import Device
from greeclimate.device_info import DeviceInfo
from greeclimate.network_helper import Props
from unittest.mock import patch


class GreeDeviceTestCase(unittest.IsolatedAsyncioTestCase):

    @patch("greeclimate.network_helper.search_devices")
    @patch("greeclimate.network_helper.bind_device")
    async def testShouldReturnKeyWhenBoundToDevice(self, mock_bind, mock_search):
        # DeviceInfo("192.168.1.29", 7000, "f4911e7aca59", "1e7aca59")

        mock_search.return_value = [
            DeviceInfo("1.1.1.0", "7000", "aabbcc001122", "MockDevice1")
        ]
        mock_bind.return_value = "St8Vw1Yz4Bc7Ef0H"

        """ The only way to get the key through binding is by scanning first
        """
        gree = GreeClimate()
        devices = await gree.search_devices()
        device = Device(devices[0])
        await device.bind()

        self.assertIsNotNone(device)
        self.assertEqual(device.device_key, "St8Vw1Yz4Bc7Ef0H")

    @patch("greeclimate.network_helper.search_devices")
    @patch("greeclimate.network_helper.bind_device")
    async def testShouldTimeoutWhenNoResponse(self, mock_bind, mock_search):

        mock_search.return_value = [
            DeviceInfo("1.1.1.0", "7000", "aabbcc001122", "MockDevice1")
        ]
        mock_bind.side_effect = socket.timeout

        """ The only way to get the key through binding is by scanning first
        """
        gree = GreeClimate()
        devices = await gree.search_devices()
        device = Device(devices[0])

        with self.assertRaises(socket.timeout):
            await device.bind()

        self.assertIsNotNone(device)
        self.assertIsNone(device.device_key)


class GreeDeviceStateTestCase(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self._device = Device(DeviceInfo("192.168.1.29", 7000, "f4911e7aca59", "1e7aca59"))
        await self._device.bind(key="St8Vw1Yz4Bc7Ef0H")

    def getMockState(self):
        return {"Pow": 1, "Mod": 3, "SetTem": 25, "TemUn": 0, "WdSpd": 0,
                "Air": 0, "Blo": 0, "Health": 0, "SwhSlp": 0, "Lig": 1,
                "SwingLfRig": 1, "SwUpDn": 1, "Quiet": 0, "Tur": 0,
                "StHt": 0, "SvSt": 0, "TemRec": 0, "HeatCoolType": 0}

    def getMockStateOff(self):
        return {"Pow": 0, "Mod": 0, "SetTem": 0, "TemUn": 0, "WdSpd": 0,
                "Air": 0, "Blo": 0, "Health": 0, "SwhSlp": 0, "Lig": 0,
                "SwingLfRig": 0, "SwUpDn": 0, "Quiet": 0, "Tur": 0,
                "StHt": 0, "SvSt": 0, "TemRec": 0, "HeatCoolType": 0}

    def getMockStateOn(self):
        return {"Pow": 1, "Mod": 1, "SetTem": 1, "TemUn": 1, "WdSpd": 1,
                "Air": 1, "Blo": 1, "Health": 1, "SwhSlp": 1, "Lig": 1,
                "SwingLfRig": 1, "SwUpDn": 1, "Quiet": 1, "Tur": 1,
                "StHt": 1, "SvSt": 1, "TemRec": 0, "HeatCoolType": 0}

    @patch("greeclimate.network_helper.request_state")
    async def testShouldUpdatePropertiesWhenRequested(self, mock_request):
        mock_request.return_value = self.getMockState()

        for p in Props:
            self.assertIsNone(self._device.get_property(p))

        await self._device.update_state()

        for p in Props:
            self.assertIsNotNone(self._device.get_property(p), f"Property {p} was unexpectedly None")
            self.assertEqual(self._device.get_property(p), self.getMockState()[p.value])

    @patch("greeclimate.network_helper.send_state")
    async def testShouldUpdatePropertiesWhenSet(self, mock_request):

        for p in Props:
            self.assertIsNone(self._device.get_property(p))

        self._device.power = True
        self._device.mode = 1
        self._device.target_temperature = 1
        self._device.temperature_units = 1
        self._device.fan_speed = 1
        self._device.fresh_air = True
        self._device.xfan = True
        self._device.anion = True
        self._device.sleep = True
        self._device.light = True
        self._device.horizontal_swing = 1
        self._device.vertical_swing = 1
        self._device.quiet = True
        self._device.turbo = True
        self._device.steady_heat = True
        self._device.power_save = True

        for p in Props:
            if p not in (Props.TEMP_BIT, Props.UNKNOWN_HEATCOOLTYPE):
                self.assertIsNotNone(self._device.get_property(p), f"Property {p} was unexpectedly None")
                self.assertEqual(self._device.get_property(p), self.getMockStateOn()[p.value])
